use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007004
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Alders
Alders'
Bieron
CIDR
DB
DROLSKY
DROLSKY's
Dave
Eilam
Eilam's
Florian
Fowler
GeoIP
Greg
Inc
Jan
MMDB
Mark
MaxMind
MaxMind's
Olaf
Oschwald
Oschwald's
PayPal
Processor
Ragwitz
Rolsky
Rolsky's
Serializer
TW
Teredo
Tree
Util
VisualizeTree
Writer
docs
drolsky
en
goschwald
jbieron
lib
mfowler
oalders
rafl
routable
tjmather
toplevel
uint
zh
